/*
 * libgsmat: An implementation of GSM 07.07
 *
 * Written by mark.liu <mark.liu@openvox.cn>
 *
 * Copyright (C) 2009-2010 OpenVox
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include "libgsmat.h"
#include "gsm_internal.h"
#include "em200.h"

static int em200_trim(char *in, char *out, int len) {
    int i=0;
    int j=0;

    for (i = 0; i < len; i++) {
        if ((in[i] != '\r') && (in[i] != '\n') ) {
            out[j++] = in[i];
        }
    }
    out[j++] = 0x0;
    return j;
}

static int em200_transmit(struct gsm_modul *gsm, char *data, int len) {
    int res;

    if (gsm->debug) {
        gsm_message(gsm, "> %s\n", data);
    }
    em200_trim(data, gsm->lastcmd, len);
    res = write(gsm->fd, data, len+2);
    if (res != (len+2)) {
        gsm_error(gsm, "Short write: %d/%d (%s) on span %d\n", res, len+2, strerror(errno), gsm->span);
        return -1;
    }
    return 0;
}

int em200_send(struct gsm_modul *gsm, char *data) {
    int len = 0;

    if (!data) {
        return -1;
    }
    len = strlen(data);
    return em200_transmit(gsm, data, len);
}

void em200_send_pin(struct gsm_modul *gsm, char *pin) {
    char buf[256];
    memset(buf, 0x0, sizeof(buf));
    if (gsm->state == GSM_STATE_SIM_PIN_REQ) {
        snprintf(buf, sizeof(buf) - 1, "AT+CPIN=\"%s\"\r", pin);
        em200_send(gsm, buf);
    }
}

static int em200_switch_state(struct gsm_modul *gsm, int state, char *next_command)
{
    gsm->state = state;
    if (next_command) {
        em200_send(gsm, next_command);
    }
    return 0;
}

static int em200_switch_uim_state(struct gsm_modul *gsm, int state, char *next_command)
{
    gsm->sim_state = state;
    if (next_command) {
        em200_send(gsm, next_command);
    }

    return 0;
}

static void em200_error_hard(void *data) {
    struct gsm_modul *gsm = (struct gsm_modul *)data;
    gsm->restart_timer = 0;
    gsm->state = GSM_STATE_ERROR;
    gsm->ev.e = GSM_EVENT_ERROR;
    gsm->ev.error.hard = 1;
    gsm->ev.error.cause = 44; /* AST_CAUSE_REQUESTED_CHAN_UNAVAIL */
    gsm->schedev = 1;
}

static void em200_get_coverage(struct gsm_modul *gsm, char *h) {
    int coverage;
    int ber;

    if (gsm_compare(h,"+CSQ:")) {
        sscanf(h, "+CSQ: %d, %d", &coverage, &ber);
        if ((coverage) && (coverage != 99)) {
            gsm->coverage = coverage;
        } else {
            gsm->coverage = -1;
        }
    }
}

static void em200_get_status(struct gsm_modul *gsm, char *h) {
    int status;

    if (gsm_compare(h,"%STATE:")) {
        sscanf(h, "%%STATE: %d", &status);
        gsm->status = status;
    }
}

static void em200_request_coverage(void *data) {
    struct gsm_modul *gsm = (struct gsm_modul *)data;
    if (gsm->state == GSM_STATE_READY) {
        em200_send(gsm, "AT+CSQ?\r");
    }
}

int em200_request_status(struct gsm_modul *gsm) {
    char buf[256];

    if (gsm->state <= GSM_STATE_READY) {
        memset(buf, 0x0, sizeof(buf));
        snprintf(buf, sizeof(buf) - 1, "AT+CSQ?\r");
        em200_switch_state(gsm, GSM_STATE_STATUS_REQ, buf);
        return 0;
    } else {
        gsm_error(gsm, "cannot request GSM status, module is in use\n");
        return -1;
    }

    return -1;
}
void em200_start(struct gsm_modul *gsm) {
    char buf[256];
    memset(buf, 0x0, sizeof(buf));
    snprintf(buf, sizeof(buf) - 1, "ATZ\r");
    em200_switch_state(gsm, GSM_STATE_POWER_ON, buf);
}

void em200_restart(void *data) {
    struct gsm_modul *gsm = (struct gsm_modul *)data;
    if (gsm->debug) {
        gsm_message(gsm, "restarting\n");
    }
    gsm->dchanup = 0;
    gsm->state = GSM_STATE_POWER_ON;
    gsm->restart_timer = gsm_schedule_event(gsm, 10000, em200_error_hard, gsm);
    em200_start(gsm);
}

int em200_answer(struct gsm_modul *gsm) {
    if (gsm->state == GSM_STATE_RINGING) {
        em200_send(gsm, "AT$QCCAV\r");
    }
    return 0;
}

int em200_dial(struct gsm_modul *gsm, int clir, char *called) {
    if (gsm->state == GSM_STATE_READY) {
        gsm->hungup = 0;
        if (clir > 0) {
            clir = 1;
        }
        gsm->clir = clir;
        strncpy(gsm->cpn, called, sizeof(gsm->cpn) - 1);
        em200_switch_state(gsm, GSM_STATE_CALL_INIT, "AT+CLIP=1\r");
        return 0;
    }
    gsm_error(gsm, "Cannot dial when in state %d\n", gsm->state);
    return -1;
}

int em200_sms_send_text(struct gsm_modul *gsm, char *destination, char *msg) {
    char buf[1024];
    if (gsm->state == GSM_STATE_READY) {
        gsm->smslen = sms_set_str(msg, buf);
        memcpy(gsm->msgbuf, buf, 1024);
        memset(buf, 0x0, sizeof(buf));
        snprintf(buf, sizeof(buf) - 1, "AT%%CMGS=\"%s\"\r", destination);
        em200_switch_state(gsm, GSM_STATE_SM_SEND_REQ, buf);
        return 0;
    }
    gsm_error(gsm, "Cannot send SM when in state %d\n", gsm->state);
    return -1;
}

int em200_hangup(struct gsm_modul *gsm) {
    char buf[256];
    gsm->cpn[0] = '\0';
    memset(buf, 0x0, sizeof(buf));
    snprintf(buf, sizeof(buf) - 1, "AT+CHV\r");
    if (!gsm->hungup && ((gsm->state == GSM_STATE_CALL_PRESENT) || (gsm->state == GSM_STATE_CALL_ACTIVE) || (gsm->state == GSM_STATE_RINGING))) {
        gsm->state = GSM_STATE_HANGUP_REQ;
        em200_transmit(gsm, buf, strlen(buf));
        if (!gsm->restart_timer) {
            gsm->restart_timer = gsm_schedule_event(gsm, 3000, em200_restart, gsm);
        }
    }
    return 0;
}

static int em200_strip(struct gsm_modul *gsm, char *in, char *out, int len) {
    int i     = 0;
    int skip  = 0;
    char *tmp = NULL;

    if ((len > 0) && ((gsm->sanidx + len) < sizeof(gsm->sanbuf))) {
        memcpy(gsm->sanbuf + gsm->sanidx, in, len);
        gsm->sanidx += len;
        gsm->sanbuf[gsm->sanidx] = 0x0;
    }

    while ((gsm->sanbuf[i] == '\r') || (gsm->sanbuf[i] == '\n')) {
        i++;
    }
    skip = i;

    if (*(gsm->sanbuf + skip + 0) == '%' &&
        *(gsm->sanbuf + skip + 1) == 'C' &&
        *(gsm->sanbuf + skip + 2) == 'M' &&
        *(gsm->sanbuf + skip + 3) == 'T' &&
        *(gsm->sanbuf + skip + 4) == ':'
        ) {
        memcpy(out, (gsm->sanbuf + skip), 1024);
        return 1;
    } else {
        tmp = memchr(gsm->sanbuf + skip, '\r', gsm->sanidx);
        if (tmp) {
            i = tmp - (gsm->sanbuf + skip);
            memcpy(out, (gsm->sanbuf + skip), i);
            out[i] = 0x0;
            memmove(gsm->sanbuf, gsm->sanbuf + skip + i, sizeof(gsm->sanbuf) - i - skip);
            gsm->sanidx -= i + skip;

            return i;
        } else {
            if (gsm->state == GSM_STATE_SM_SEND_REQ) {
                if ((gsm->sanidx - skip == 1) && (gsm->sanbuf[gsm->sanidx - 1] == '>')) {
                    out[0] = '>';
                    out[1] = 0x0;
                    gsm->sanidx = 0;
                    return 1;
                }
            }
        }

    }

    return 0;
}

gsm_event *em200_receive(struct gsm_modul *gsm, char *data, int len) {
    int res;
    char buf[1024];
    char buf2[4096];
    char callerid[256];
    char *a = NULL;

    res = em200_strip(gsm, data, buf2, len);
    if (!res) {
        return NULL;
    }

    if (gsm->debug) {
        gsm_message(gsm, "< %s , %d >\n",buf2, res);
    }

    if (gsm_compare(gsm->lastcmd, "AT+CSQ?") && gsm->state != GSM_STATE_SM_RECEIVING) {
        if ((gsm_compare(buf2, "+CSQ:")) && (gsm->state >=  GSM_STATE_READY)){
            em200_get_coverage(gsm, buf2);
            if (gsm->coverage == -1) {
                em200_start(gsm);
                if (gsm->dchanup) {
                    gsm->dchanup = 0;
    //              gsm->state = GSM_STATE_READY;
                    gsm->ev.e = GSM_EVENT_DCHAN_DOWN;
                    return &gsm->ev;
                }
            }
        } else if (gsm_compare(buf2, "OK")) {
            if (gsm->coverage >= 1) {
                usleep(1000);
                em200_send(gsm, "AT%STATE\r");
                if (!gsm->dchanup) {
                    gsm->dchanup = 1;
                    em200_switch_state(gsm, GSM_STATE_READY, "AT+CLIP=1\r");
                    gsm->ev.e = GSM_EVENT_DCHAN_UP;
                    return &gsm->ev;
                }
            } else {
                em200_start(gsm);
                if (gsm->dchanup) {
                    gsm->dchanup = 0;
    //              gsm->state = GSM_STATE_READY;
                    gsm->ev.e = GSM_EVENT_DCHAN_DOWN;
                    return &gsm->ev;
                }
            }
        } else {
//          gsm_error(gsm, "!%s!\n", buf2);
        }
    }

    if (gsm_compare(gsm->lastcmd, "AT%STATE") && gsm->state != GSM_STATE_SM_RECEIVING) {
        if((gsm_compare(buf2, "%STATE:")) && (gsm->state >=  GSM_STATE_READY)){
            em200_get_status(gsm, buf2);
        } else if (gsm_compare(buf2, "OK")) {
            if (gsm->status != 0) {
                gsm_schedule_event(gsm, 10000, em200_request_coverage, gsm);
                if (!gsm->dchanup) {
                    gsm->dchanup = 1;
                    usleep(1000);
                    em200_switch_state(gsm, GSM_STATE_READY, "AT+CLIP=1\r");
                    gsm->ev.e = GSM_EVENT_DCHAN_UP;
                    return &gsm->ev;
                }
            } else {
                usleep(1000);
                em200_start(gsm);
                if (gsm->dchanup) {
                    gsm->dchanup = 0;
                    gsm->ev.e = GSM_EVENT_DCHAN_DOWN;
                    return &gsm->ev;
                }
            }
        } else {
//          gsm_error(gsm, "!%s!\n", buf2);
        }
    }

    switch (gsm->state) {
    case GSM_STATE_POWER_ON:
        if (gsm_compare(buf2, "OK")) {
            if (gsm_compare(gsm->lastcmd, "ATZ")) {
                sleep(1);
                em200_switch_state(gsm, GSM_STATE_INIT_0, "ATE0\r");
            }
            if (gsm->restart_timer) {
                gsm_schedule_del(gsm, gsm->restart_timer);
                gsm->restart_timer = 0;
            }
        } else if (gsm_compare(buf2, "ATZ")) {
            if (gsm->restart_timer) {
                gsm_schedule_del(gsm, gsm->restart_timer);
                gsm->restart_timer = 0;
            }
        } else if (gsm_compare(buf2, "%STATE")){
        } else {
            gsm_error(gsm, DEBUGFMT "!%s!\n", DEBUGARGS, buf2);
        }
        break;
    case GSM_STATE_INIT_0:
      if (gsm_compare(buf2, "OK")) {
            if (gsm_compare(gsm->lastcmd, "ATE0")) {
                em200_switch_state(gsm, GSM_STATE_INIT_1, "AT$HCTCM=0\r");
            }
            if (gsm->restart_timer) {
                gsm_schedule_del(gsm, gsm->restart_timer);
                gsm->restart_timer = 0;
            }
        } else if (gsm_compare(buf2, "ATE0")) {
            if (gsm->restart_timer) {
                gsm_schedule_del(gsm, gsm->restart_timer);
                gsm->restart_timer = 0;
            }
        } else if (gsm_compare(buf2, "NO CARRIER") || gsm_compare(buf2, "ERROR")) {
                gsm->restart_timer = gsm_schedule_event(gsm, 1000, em200_restart, gsm);
        } else {
            gsm_error(gsm, DEBUGFMT "!%s!\n", DEBUGARGS, buf2);
        }
        break;
    case GSM_STATE_INIT_1:
        if (gsm_compare(buf2, "OK")) {
            sleep(1); /* To show the network status */
            em200_switch_state(gsm, GSM_STATE_INIT_2, "AT+CMEE=2\r");
        } else if (gsm_compare(buf2, "$HCTCM: 0")){
        } else if (gsm_compare(buf2, "ATE0")){
        } else {
            gsm_error(gsm, DEBUGFMT "!%s!\n", DEBUGARGS, buf2);
        }
        break;
    case GSM_STATE_INIT_2:
        if (gsm_compare(buf2, "OK")) {
            em200_switch_state(gsm, GSM_STATE_INIT_3, "AT+CPIN?\r");
        } else {
            gsm_error(gsm, DEBUGFMT "!%s!\n", DEBUGARGS, buf2);
        }
        break;
    case GSM_STATE_INIT_3:
        if (gsm_compare(buf2, "+CPIN:READY")) {
            em200_switch_uim_state(gsm, GSM_STATE_SIM_READY, NULL);
        } else if (gsm_compare(buf2, "+CPIN: SIM PIN")) {
            em200_switch_uim_state(gsm, GSM_STATE_SIM_PIN_REQ, NULL);
        } else if (gsm_compare(buf2, "OK")) {
            switch(gsm->sim_state) {
            case GSM_STATE_SIM_READY: /* sim pin ready */
                em200_switch_state(gsm, GSM_STATE_SIM_READY, "AT+CVER\r");
                break;
            case GSM_STATE_SIM_PIN_REQ: /* sim pin required */
                em200_switch_state(gsm, GSM_STATE_SIM_PIN_REQ, NULL);
                gsm->ev.e = GSM_EVENT_PIN_REQUIRED;
                return &gsm->ev;
                break;
            }
        } else {
            gsm_error(gsm, DEBUGFMT "!%s!\n", DEBUGARGS, buf2);
            if (gsm_compare(buf2, "+CME ERROR:UIM not inserted")) {
                em200_start(gsm);
            }
        }
        break;
    case GSM_STATE_SIM_PIN_REQ:
        if (gsm_compare(buf2, "OK")) {
            em200_switch_state(gsm, GSM_STATE_SIM_READY, "AT+CVER\r");
        } else if (gsm_compare(buf2, "+CME ERROR: SIM not inserted")) {
            /* re-enter SIM if neccessary */
            gsm->state = GSM_STATE_SIM_PIN_REQ;
            gsm->ev.e = GSM_EVENT_PIN_REQUIRED;
            return &gsm->ev;
        } else {
            gsm_error(gsm, DEBUGFMT "!%s!\n", DEBUGARGS, buf2);
        }
        break;
    case GSM_STATE_SIM_READY:
        if (gsm_compare(buf2, "OK")) { /* Get unsolicited GSM registration reports */
            em200_switch_state(gsm, GSM_STATE_READY, "AT%STATE\r");
        } else if (gsm_compare(buf2, "+CVER:")){
        } else {
            gsm_error(gsm, DEBUGFMT "!%s!\n", DEBUGARGS, buf2);
        }
        break;
    case GSM_STATE_READY:
        if (gsm_compare(gsm->lastcmd, "AT+CLIP=1")) { /* Calling Line Identification. Enable CLI presentation */
            usleep(1000);
            em200_send(gsm, "AT%SMSSS=1,0,6,0\r");
        } else if (gsm_compare(gsm->lastcmd, "AT%SMSSS=1,0,6,0")) { // ASCII
            usleep(1000);
            em200_send(gsm, "AT+CNMI=1,2,0,1,0\r");
        } else if (gsm_compare(gsm->lastcmd, "AT+CNMI=1,2,0,1,0")) {
            usleep(1000);
            em200_send(gsm, "AT+CLVL=4\r");
        } else if (gsm_compare(gsm->lastcmd, "AT+CLVL=4")) {
            usleep(1000);
            em200_send(gsm, "AT%MICV=4\r");
        } else if (gsm_compare(gsm->lastcmd, "AT%MICV=4")) {
            usleep(1000);
            em200_send(gsm, "AT%SDEV=0\r");
        } else if (gsm_compare(gsm->lastcmd, "AT%SDEV=0")) {
            usleep(1000);
            em200_send(gsm, "AT%SETEC=0\r"); /* disable echo cancellation */
        } else if (gsm_compare(gsm->lastcmd, "AT%SETEC=0")) {
            usleep(1000);
             /* disable echo cancellation */
        }

        callerid[0] = '\0';
        gsm->ev.ring.callingnum[0] = '0';
        if (gsm_compare(buf2, "+CLIP:")) { /* ring */
            gsm->state = GSM_STATE_RINGING;
            if (sscanf(buf2, "+CLIP:%s,0", callerid) == 1) {
                a = strchr(callerid, '"');
                if (a) {
                    *a = '\0';
                }
            }
            gsm->hungup = 0;
            if (!strlen(callerid)) {
                strncpy(gsm->ev.ring.callingnum, "", sizeof(gsm->ev.ring.callingnum) - 1);
            } else {
                strncpy(gsm->ev.ring.callingnum, callerid, sizeof(gsm->ev.ring.callingnum) - 1);
            }
            gsm->ev.e = GSM_EVENT_RING;
            return &gsm->ev;
        } else if (gsm_compare(buf2, "%CMT:")) { /* SMS received */
            //memset(gsm->msgbuf, 0, sizeof(gsm->msgbuf)); /* get SMS */
            memcpy(gsm->msgbuf, buf2, sizeof(gsm->msgbuf) - 1); /* get SMS */
            gsm->state = GSM_STATE_SM_RECEIVING;
        } else if (gsm_compare(buf2, "ERROR")) {
            /* nothing */
            if (!gsm_compare(gsm->lastcmd, "AT+CHV")) {
                gsm->state = GSM_STATE_READY;
                gsm->ev.e = GSM_EVENT_HANGUP;
                gsm->ev.hangup.cause = 16;
                return &gsm->ev;
            }
        } else if (gsm_compare(buf2, "NO CARRIER")) {
            if (!gsm_compare(gsm->lastcmd, "AT+CHV")) {
                gsm->state = GSM_STATE_READY;
                gsm->ev.e = GSM_EVENT_HANGUP;
                gsm->ev.hangup.cause = 16;
                return &gsm->ev;
            }
        } else if (gsm_compare(buf2, "%CEND: 10") || gsm_compare(buf2, "%CEND: 9")) {
            if (!gsm_compare(gsm->lastcmd, "AT+CHV")) {
                gsm->state = GSM_STATE_READY;
                gsm->ev.e = GSM_EVENT_HANGUP;
                gsm->ev.hangup.cause = 16;
                return &gsm->ev;
            }
        } else if (gsm_compare(buf2, "OK")) {
            if (gsm_compare(gsm->lastcmd, "AT+CDV")) {
                gsm->state = GSM_STATE_CALL_PRESENT;
            } else if (gsm_compare(gsm->lastcmd, "AT+CMGF=1")){
            } else if (gsm_compare(gsm->lastcmd, "AT+CLVL=4")){
            } else if (gsm_compare(gsm->lastcmd, "AT%MICV=4")){
            } else if (gsm_compare(gsm->lastcmd, "AT%SDEV=0")){
            } else if (gsm_compare(gsm->lastcmd, "AT%SETEC=0")){
                /* echo */
            }
        } else if (gsm_compare(buf2, "+CRING: 0")) {
        } else if (gsm_compare(buf2, "$HVOL:")) {
        } else if (gsm_compare(buf2, "+CSQ:")) {
        } else if (gsm_compare(buf2, "%STATE:")) {
        } else if (gsm_compare(buf2, "%CMGSS")) { // sms send failure
        } else if (gsm_compare(buf2, "%CMGSF")) { // sms send error
            gsm_error(gsm, "Sending SMS error.\n");
        } else {
            gsm_error(gsm, DEBUGFMT "!%s!\n", DEBUGARGS, buf2);
        }
        break;
    case GSM_STATE_RINGING:
        if (gsm_compare(buf2, "NO CARRIER")) { /* hang up */
            gsm->state = GSM_STATE_READY;
            gsm->ev.e = GSM_EVENT_HANGUP;
            gsm->ev.hangup.cause = 16;
            return &gsm->ev;
        } else if (gsm_compare(buf2, "OK")) {
            if (gsm_compare(gsm->lastcmd, "AT$QCCAV")) { /* if answer */
                gsm->state = GSM_STATE_CALL_ACTIVE;

            }
        } else if (gsm_compare(buf2, "%CCNT: 0")) { // before AT$QCCAV
        } else {
            gsm_error(gsm, DEBUGFMT "!%s!\n", DEBUGARGS, buf2);
        }
        break;
    case GSM_STATE_CALL_ACTIVE:
        if ((gsm_compare(buf2, "%CEND: 6, 0"))) {
            /* Received a Release from Base Station (This is a normal call termination) */
            em200_switch_state(gsm, GSM_STATE_READY, "AT%STATE\r");
            gsm->ev.e = GSM_EVENT_HANGUP;
            gsm->ev.hangup.cause = 16;
            return &gsm->ev;
        } else if (gsm_compare(buf2, "OK")) {
            if (gsm_compare(gsm->lastcmd, "AT+CHV")) {
                em200_switch_state(gsm, GSM_STATE_READY, "AT%STATE\r");
            }
        } else if (gsm_compare(buf2, "%CCNT: 0")) { /* Active */
        } else {
            gsm_error(gsm, DEBUGFMT "!%s!\n", DEBUGARGS, buf2);
        }
        break;
    case GSM_STATE_CALL_INIT:
        if (gsm_compare(buf2, "OK")) {
            if (gsm_compare(gsm->lastcmd, "AT+CLIP=1")) {
                gsm->lastcmd[0] = '\0';
                if (strlen(gsm->cpn)) {
                    memset(buf, 0x0, sizeof(buf));
                    snprintf(buf, sizeof(buf) - 1, "AT+CDV%s\r", gsm->cpn);
                    em200_switch_state(gsm, GSM_STATE_CALL_PRESENT, buf);
                } else {
                    em200_switch_state(gsm, GSM_STATE_READY, "AT%STATE\r");
                }
            }
        }
        break;
    case GSM_STATE_CALL_PRESENT:
        /* Connection Failure - NO CARRIER or BUSY or NO ANSWER */
        if (gsm_compare(buf2, "+CME ERROR")) {
            /* modem is in weird state, reboot */
            gsm_error(gsm, "error while dialing (%s)\n", buf2);
            gsm->state = GSM_STATE_ERROR;
            gsm->ev.e = GSM_EVENT_ERROR;
            gsm->ev.error.hard = 1;
            gsm->ev.error.cause = 44; /* AST_CAUSE_REQUESTED_CHAN_UNAVAIL */
            strncpy(gsm->ev.error.err, buf2, sizeof(gsm->ev.error.err) - 1);
            return &gsm->ev;
        } else if (gsm_compare(buf2, "%CEND: 2") || gsm_compare(buf2, "NO CARRIER")) { /* no service, no network */
            if (!gsm_compare(gsm->lastcmd, "AT+CHV")) {
                em200_switch_state(gsm, GSM_STATE_READY, "AT%STATE\r");
                gsm->ev.e = GSM_EVENT_HANGUP;
                gsm->ev.hangup.cause = 16;
                return &gsm->ev;
            }
            em200_switch_state(gsm, GSM_STATE_READY, "AT%STATE\r");
        } else if (gsm_compare(buf2, "%CCNT: 0")) { /* Alerting (MO call) */
            gsm->ev.e = GSM_EVENT_ALERTING;
            em200_send(gsm, "ATE0\r");
            return &gsm->ev;
        } else if (gsm_compare(buf2, "OK")) {
            if (gsm_compare(gsm->lastcmd, "AT+CDV")) {
                gsm->lastcmd[0] = '\0';
            } else if (gsm_compare(gsm->lastcmd, "ATE0")){
                usleep(1000);
                em200_send(gsm, "AT%MICV=4\r");
            } else if (gsm_compare(gsm->lastcmd, "AT%MICV=4")){
                usleep(1000);
                em200_send(gsm, "AT+CLVL=4\r");
            } else if (gsm_compare(gsm->lastcmd, "AT+CLVL=4")){
                gsm->state = GSM_STATE_CALL_ACTIVE;
                gsm->ev.e = GSM_EVENT_ANSWER; /* Answer */
                return &gsm->ev;
            }
        } else if (gsm_compare(buf2, "%CEND: 6, 0")) {
            em200_switch_state(gsm, GSM_STATE_READY, "AT%STATE\r");
            gsm->ev.e = GSM_EVENT_HANGUP;
            gsm->ev.hangup.cause = 16;
            return &gsm->ev;
        } else if (gsm_compare(buf2, "%CORG:")) {
        } else if (gsm_compare(buf2, "$HVOL")) {
        } else if (gsm_compare(buf2, "NO CARRIER")) {
        } else {
            gsm_error(gsm, DEBUGFMT "!%s!\n", DEBUGARGS, buf2);
        }
        break;
    case GSM_STATE_SM_SEND_REQ:
        if (gsm_compare(buf2, ">")) {
            em200_transmit(gsm, gsm->msgbuf, gsm->smslen);
            memset(gsm->msgbuf, 0x0, sizeof(gsm->msgbuf));
            snprintf(gsm->msgbuf, sizeof(gsm->msgbuf) - 1, "%c", 0x1A); // send sms endl
            em200_switch_state(gsm, GSM_STATE_SM_SENDING, gsm->msgbuf);
        } else if (gsm_compare(buf2, "+CMS ERROR")) {
            gsm_error(gsm, "error sending SMS on span %d (%s)\n", gsm->span,  buf2);
            em200_switch_state(gsm, GSM_STATE_READY, "AT%STATE\r");
        } else if (gsm_compare(buf2, "+CME ERROR")) {
            gsm_error(gsm, "error sending SMS on span %d (%s)\n", gsm->span,  buf2);
            em200_switch_state(gsm, GSM_STATE_READY, "AT%STATE\r");
        }
        break;
    case GSM_STATE_SM_SENDING:
        if (gsm_compare(buf2, "%CMGSS:")) {
//          em200_switch_state(gsm, GSM_STATE_READY, "AT+CMGF=1\r"); /* only support text mode */
            gsm_message(gsm, "SMS sent successfully on span %d.\n", gsm->span);
            em200_switch_state(gsm, GSM_STATE_READY, "AT%STATE\r");
        } else if (gsm_compare(buf2, "+CMS ERROR:")) {
            gsm_error(gsm, "error sending sms (%s)\n", buf2);
            em200_switch_state(gsm, GSM_STATE_READY, "AT%STATE\r");
        }
        break;
    case GSM_STATE_SM_RECEIVING:
        if (!gsm_compare(gsm->lastcmd, "AT+CNMA")) {
            em200_send(gsm, "AT+CNMA\r"); /* New Message Acknowledgment */
        } else {
            gsm->lastcmd[0] = '\0';
            gsm->sanidx = 0;
            memset(gsm->sanbuf, 0, 4096);
            em200_switch_state(gsm, GSM_STATE_READY, "AT+CSQ?\r");
            gsm->ev.e = GSM_EVENT_SM_RECEIVED;
            if (!gsm_text2sm_event(gsm, gsm->msgbuf)) {
                return &gsm->ev;
            } 
            if(gsm_compare(buf2, "ERROR")|| gsm_compare(buf2, "+CMS ERROR") || gsm_compare(buf2, "NO CARRIER")){
                    em200_start(gsm);
            }
        }
        break;
    case GSM_STATE_HANGUP_REQ:
        if (gsm_compare(buf2, "OK")) {
            em200_switch_state(gsm, GSM_STATE_READY, "AT%STATE\r");
            if (gsm->restart_timer) {
                gsm_schedule_del(gsm, gsm->restart_timer);
                gsm->restart_timer = 0;
            }
        } else if (gsm_compare(buf2, "%CEND:")) {
        } else {
            gsm_error(gsm, DEBUGFMT "!%s!\n", DEBUGARGS, buf2);
        }
        break;
    case GSM_STATE_STATUS_REQ:
        if (gsm_compare(buf2, "+CSQ:")){
            if (gsm->restart_timer) {
                gsm_schedule_del(gsm, gsm->restart_timer);
                gsm->restart_timer = 0;
            }
            gsm->state = GSM_STATE_READY;
            gsm_message(gsm, "signal strength on span %d (%s) [min 0, max 31]\n", gsm->span, buf2);
        }
        break;

    }

    if ((gsm_compare(buf2, "NO CARRIER")) && (gsm->state >  GSM_STATE_READY)){
        gsm->state = GSM_STATE_READY;
        gsm->ev.e = GSM_EVENT_HANGUP;
        gsm->ev.hangup.cause = 16;
        return &gsm->ev;
    }

    if ((gsm_compare(buf2, "NO DIALTONE")) && (gsm->state >  GSM_STATE_READY)){
        gsm->state = GSM_STATE_READY;
        gsm->ev.e = GSM_EVENT_HANGUP;
        gsm->ev.hangup.cause = 16;
        return &gsm->ev;
    }

    if ((gsm_compare(buf2, "BUSY")) && (gsm->state >  GSM_STATE_READY)){
        gsm->state = GSM_STATE_READY;
        gsm->ev.e = GSM_EVENT_HANGUP;
        gsm->ev.hangup.cause = 16;
        return &gsm->ev;
    }


    return NULL;
}
