/*
 * libgsmat: An implementation of GSM 07.07
 *
 * Written by Klaus-Peter Junghanns <support@junghanns.net>
 *
 * Copyright (C) 2005, 2006, 2007 Junghanns.NET GmbH
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include "libgsmat.h"
#include "gsm_internal.h"
#include "g20.h"

int g20_trim(char *data, char *out, int len);

static int g20_transmit(struct gsm_modul *gsm, char *data, int len) {
    int res;

    if (gsm->debug) {
        gsm_message(gsm, "> %s\n", data);
    }
    g20_trim(data, gsm->lastcmd, len);
    res = write(gsm->fd, data, len+2);
    if (res != (len+2)) {
        gsm_error(gsm, "Short write: %d/%d (%s) on span %d\n", res, len+2, strerror(errno), gsm->span);
        return -1;
    }
    return 0;
}

int g20_send(struct gsm_modul *gsm, char *data) {
    int len = 0;
    
    if (!data) {
        return -1;
    }
    len = strlen(data);
    return g20_transmit(gsm, data, len);
}

void g20_send_pin(struct gsm_modul *gsm, char *pin) {
    char buf[256];
    memset(buf, 0x0, sizeof(buf));
    if (gsm->state == GSM_STATE_SIM_PIN_REQ) {
        snprintf(buf, sizeof(buf) - 1, "AT+CPIN=\"%s\"\r", pin);
        g20_transmit(gsm, buf, strlen(buf)); 
    }
}

static int g20_switch_state(struct gsm_modul *gsm, int state, char *next_command) 
{
    gsm->state = state;
    if (next_command) {
        g20_send(gsm, next_command);
    }
    return 0;
}

static int g20_switch_sim_state(struct gsm_modul *gsm, int state, char *next_command) 
{
    gsm->sim_state = state;
    if (next_command) {
        g20_send(gsm, next_command);
    }

    return 0;
}

static void g20_error_hard(void *data) {
    struct gsm_modul *gsm = (struct gsm_modul *)data;
    gsm->restart_timer = 0;
    gsm->state = GSM_STATE_ERROR;
    gsm->ev.e = GSM_EVENT_ERROR;
    gsm->ev.error.hard = 1;
    gsm->ev.error.cause = 44; /* AST_CAUSE_REQUESTED_CHAN_UNAVAIL */
    gsm->schedev = 1;
}

void g20_start(struct gsm_modul *gsm) {
    char buf[256];
    memset(buf, 0x0, sizeof(buf));
    snprintf(buf, sizeof(buf) - 1, "ATZ\r");
    g20_transmit(gsm, buf, strlen(buf));
}

void g20_restart(void *data) {
    struct gsm_modul *gsm = (struct gsm_modul *)data;
    if (gsm->debug) {
        gsm_message(gsm, "restarting\n");
    }
    gsm->dchanup = 0;
    gsm->state = GSM_STATE_POWER_ON;
    gsm->restart_timer = gsm_schedule_event(gsm, 10000, g20_error_hard, gsm);
    g20_start(gsm);
}

int g20_answer(struct gsm_modul *gsm) {
    if (gsm->state == GSM_STATE_RINGING) {
        g20_transmit(gsm, "ATA\r", 4); 
    }
    return 0; 
}

int g20_poweroff(struct gsm_modul *gsm) {
    g20_transmit(gsm, "ATZ\r", 4); 
    g20_transmit(gsm, "AT+MRST\r", 8); 
    return 0; 
}

int g20_dial(struct gsm_modul *gsm, int clir, char *called) {
    char buf[256];
    if (gsm->state == GSM_STATE_READY) {
        gsm->hungup = 0;
        if (clir > 0) clir = 1;
        gsm->clir = clir;
        strncpy(gsm->cpn, called, sizeof(gsm->cpn) - 1);
        memset(buf, 0x0, sizeof(buf));
        snprintf(buf, sizeof(buf) - 1, "AT+CLIR=%d\r", clir);
        g20_switch_state(gsm, GSM_STATE_CALL_INIT, buf);
        return 0;
    }
    gsm_error(gsm, "Cannot dial when in state %d\n", gsm->state);
    return -1;
}

int g20_sms_send_text(struct gsm_modul *gsm, char *destination, char *msg) {
    char buf[1024];
    if (gsm->state == GSM_STATE_READY) {
        snprintf(gsm->msgbuf, sizeof(gsm->msgbuf) - 1, "%s", msg);

        memset(buf, 0x0, sizeof(buf));
        snprintf(buf, sizeof(buf) - 1, "AT+CMGF=1\r"); /* text mode */
        g20_transmit(gsm, buf, strlen(buf)); 
        memset(buf, 0x0, sizeof(buf));
        snprintf(buf, sizeof(buf) - 1, "AT+CMGS=\"%s\"\r", destination);
        g20_switch_state(gsm, GSM_STATE_SM_SEND_REQ, buf);
        return 0;
    }
    gsm_error(gsm, "Cannot send SM when in state %d\n", gsm->state);
    return -1;
}

int g20_sms_send_pdu(struct gsm_modul *gsm, char *pdu) {
    char buf[1024];
    char smsc[3];
    int smsc_len;
    int len = 0;

    if (!pdu) return 0;
    
    len = strlen(pdu);

    /* SMSC information length */
    strncpy(smsc, pdu, sizeof(smsc) - 1);
    smsc[2] = '\0';
    smsc_len = gsm_hex2int(smsc, 2);
    len = (len / 2) - 1 - smsc_len;

    if (gsm->state == GSM_STATE_READY) {
        snprintf(gsm->msgbuf, sizeof(gsm->msgbuf) - 1, "%s", pdu);

        memset(buf, 0x0, sizeof(buf));
        snprintf(buf, sizeof(buf) - 1, "AT+CMGF=0\r"); /* pdu mode */
        g20_transmit(gsm, buf, strlen(buf)); 
        memset(buf, 0x0, sizeof(buf));
        snprintf(buf, sizeof(buf) - 1, "AT+CMGS=%d\r", len);
        g20_switch_state(gsm, GSM_STATE_SM_SEND_REQ, buf);
        return 0;
    }
    gsm_error(gsm, "Cannot send SM when in state %d\n", gsm->state);
    return -1;
}

int g20_request_status(struct gsm_modul *gsm) {
    char buf[256];
    if (gsm->state <= GSM_STATE_READY) {
    memset(buf, 0x0, sizeof(buf));
    snprintf(buf, sizeof(buf) - 1, "AT+CSQ?\r");
    gsm->state = GSM_STATE_STATUS_REQ;    
    g20_transmit(gsm, buf, strlen(buf));
    } else {
    gsm_error(gsm, "cannot request GSM status, module is in use\n");
    }
    return 0; 
}

int g20_hangup(struct gsm_modul *gsm) {
    char buf[256];
    gsm->cpn[0] = '\0';
    memset(buf, 0x0, sizeof(buf));
    snprintf(buf, sizeof(buf) - 1, "ATH\r");
    if (!gsm->hungup && ((gsm->state == GSM_STATE_CALL_PRESENT) || (gsm->state == GSM_STATE_CALL_ACTIVE) || (gsm->state == GSM_STATE_RINGING))) {
        gsm->state = GSM_STATE_HANGUP_REQ;
        g20_transmit(gsm, buf, strlen(buf));
        if (!gsm->restart_timer) {
            gsm->restart_timer = gsm_schedule_event(gsm, 3000, g20_restart, gsm);
        }
    }
    return 0; 
}


int g20_trim(char *data, char *out, int len) {
    int i=0;
    int j=0;

    for (i=0;i<len;i++) {
        if ((data[i] != '\r') && (data[i] != '\n') ) {
            out[j++] = data[i]; 
        }
    }
    out[j++] = 0x0;
    return j;
}

static int g20_san(struct gsm_modul *gsm, char *in, char *out, int len) {
    int i=0;
    int skip=0;
    char *tmp = NULL;

    if ((len > 0) && ((gsm->sanidx + len) < sizeof(gsm->sanbuf))) {
        memcpy(gsm->sanbuf + gsm->sanidx, in, len);
        gsm->sanidx += len;
        gsm->sanbuf[gsm->sanidx] = 0x0;
    }

    while ((gsm->sanbuf[i] == '\r') || (gsm->sanbuf[i] == '\n')) {
        i++;
    }
    
    skip = i;

    tmp = strchr(gsm->sanbuf + skip, '\r');
    if (tmp){
        i = tmp - (gsm->sanbuf + skip);
        memcpy(out, (gsm->sanbuf + skip), i);
        out[i] = 0x0;
        memmove(&gsm->sanbuf[0], &gsm->sanbuf[skip+i], sizeof(gsm->sanbuf) - i - skip);
        gsm->sanidx -= i + skip;
        return i;
    } else {
        if (gsm->state == GSM_STATE_SM_SEND_REQ) {
            if ((gsm->sanidx - skip == 2) && (gsm->sanbuf[gsm->sanidx - 2] == '>') && (gsm->sanbuf[gsm->sanidx - 1] == ' ')) {
                out[0] = '>';
                out[1] = ' ';
                out[2] = 0x0;
                gsm->sanidx -= 2 + skip;
                return 2;
            }
        }
    }

    return 0;
}

gsm_event *g20_receive(struct gsm_modul *gsm, char *data, int len) {
    int res;
    char buf[1024];
    char buf2[1024];
    char buf3[1024];
    char callerid[256];
    char loc_str[256];
    char *a = NULL;
    int creg1 = 0;
    int creg2 = 0;
    int creg3 = 0;
    int registered = 0;
    char tmp[256];

    res = g20_san(gsm, data, buf, len);
    if (!res) {
        return NULL;
    }
    
    g20_trim(buf, buf2, res);

    if (gsm->debug) {
        gsm_message(gsm, "< %s , %d >\n",buf2, res);
    }
    
    strncpy(buf3, buf2, sizeof(buf3) - 1);
    if (gsm_compare(buf3, "+CREG:")) {
        if (gsm_compare(gsm->lastcmd, "AT+CREG?")) {
            if (sscanf(buf3, "+CREG: %d,%d,%s", &creg1, &creg2, loc_str) == 3) {
                // gsm_message(gsm, "creg %d %s\n", creg1, loc_str);
            }
            gsm->lastcmd[0] = '\0';
        } else {
            if (gsm_compare(buf3, "+CREG: 002")) {
                /* Not registered, but the ME is currently searching for a
                   new operator to which to register.*/
                creg2 = 2;
            } else if (gsm_compare(buf3, "+CREG: 003")) {
                /* Registration denied. */
                creg2 = 3;
            } else if (sscanf(buf3, "+CREG: %d,%s", &creg1, loc_str) == 2) {
                /* unsolicited */
                // gsm_message(gsm, "creg %d %d %d\n", creg1, creg2, creg3);
                creg2 = creg1;
            }
        }

        /* <stat> : 1 Registered, home network 
                    5 Registered, roaming
         */
        registered = (creg2 == 1) || (creg2 == 5) || (creg2 == 8);
        if (registered) {
            if (!gsm->dchanup) {
                gsm->dchanup = 1;
                g20_switch_state(gsm, GSM_STATE_READY, "AT+CLIP=1\r");
                gsm->ev.e = GSM_EVENT_DCHAN_UP;
                return &gsm->ev;
            }
        } else {
            if (gsm->dchanup) {
                gsm->dchanup = 0;
                gsm->state = GSM_STATE_READY;
                gsm->ev.e = GSM_EVENT_DCHAN_DOWN;
                return &gsm->ev;
            }
        }
    }
    
    switch (gsm->state) {
    case GSM_STATE_POWER_ON:
        if (gsm_compare(buf2, "OK")) {
            if (gsm_compare(gsm->lastcmd, "ATZ")) {
                g20_switch_state(gsm, GSM_STATE_INIT_0, "AT+CMEE=2\r");
            }
            if (gsm->restart_timer) {
                gsm_schedule_del(gsm, gsm->restart_timer);
                gsm->restart_timer = 0;
            }
        } else if (gsm_compare(buf2, "ATZ")) {
            /* echo */
            if (gsm->restart_timer) {
                gsm_schedule_del(gsm, gsm->restart_timer);
                gsm->restart_timer = 0;
            }
        } else {
            gsm_error(gsm, "!%s!\n", buf2);
        } 
        break;
    case GSM_STATE_INIT_0:
        if (gsm_compare(buf2, "OK")) {
            g20_switch_state(gsm, GSM_STATE_INIT_1, "AT+MADIGITAL=1\r");
        } else if (gsm_compare(buf2, "AT+CMEE=2")) {
            /* echo */
        } else {
            gsm_error(gsm, "!%s!\n", buf2);
        } 
        break;
    case GSM_STATE_INIT_1:
        if (gsm_compare(buf2, "OK")) {
            g20_switch_state(gsm, GSM_STATE_INIT_2, "AT+CPIN?\r");
        } else if (gsm_compare(buf2, "AT+MADIGITAL=1")) {
            /* echo */
        } else {
            gsm_error(gsm, "!%s!\n", buf2);
        } 
        break;
    case GSM_STATE_INIT_2:
        if (gsm_compare(buf2, "+CPIN: READY")) {
            g20_switch_sim_state(gsm, GSM_STATE_SIM_READY, "AT+CREG=2\r");
        } else if (gsm_compare(buf2, "+CPIN: SIM PIN")) {
            g20_switch_sim_state(gsm, GSM_STATE_SIM_PIN_REQ, NULL);
        } else if (gsm_compare(buf2, "OK")) {
            switch(gsm->sim_state) {
            case GSM_STATE_SIM_READY: /* sim pin ready */
                g20_switch_state(gsm, GSM_STATE_SIM_READY, "AT+CREG=2\r");
                break;
            case GSM_STATE_SIM_PIN_REQ: /* sim pin required */
                g20_switch_state(gsm, GSM_STATE_SIM_PIN_REQ, NULL);
                gsm->ev.e = GSM_EVENT_PIN_REQUIRED;
                return &gsm->ev;
                break;
            }
        } else if (gsm_compare(buf2, "AT+CPIN?")) {
            /* echo */
        } else {
            gsm_error(gsm, "!%s!\n", buf2);
        } 
        break;
    case GSM_STATE_SIM_PIN_REQ:
        if (gsm_compare(buf2, "OK")) {
            //gsm_message(gsm, "LAST>%s<\n",gsm->lastcmd);
            /* Enables the network registration and location
               information in unsolicited reports and Read command
               +CREG:<stat>[,<lac>,<ci>].*/
            g20_switch_state(gsm, GSM_STATE_SIM_READY, "AT+CREG=2\r");
        } else if (gsm_compare(buf2, "AT+CREG=2")) {
        } else if (gsm_compare(buf2, "AT+CPIN?")) {
            /* echo */
        } else if (gsm_compare(buf2, "+CME ERROR: SIM not inserted")) {
            /* re-enter SIM if neccessary */
            gsm->state = GSM_STATE_SIM_PIN_REQ;
            gsm->ev.e = GSM_EVENT_PIN_REQUIRED;
            return &gsm->ev;
        } else {
            gsm_error(gsm, "!%s!\n", buf2);
        } 
        break;
    case GSM_STATE_SIM_READY:
        if (gsm_compare(buf2, "OK")) { /* Get unsolicited GSM registration reports */
            g20_switch_state(gsm, GSM_STATE_NET_REQ, "AT+CREG?\r");
        } else if (gsm_compare(buf2, "AT+CREG=2")) { 
            /* echo */
        } else {
            gsm_error(gsm, "!%s!\n", buf2);
        } 
        break;
    case GSM_STATE_NET_REQ:
        if (gsm_compare(buf2, "+CREG:")) {
            if (sscanf(buf2, "+CREG: %d,%d,%d", &creg1, &creg2, &creg3) == 3) {
            // gsm_message(gsm, "creg %d %d %d\n", creg1, creg2, creg3);
            // +CREG: 002,001,253d,9b22
            }
        } else {
        //    gsm_error("!%s!\n", buf2);
        } 
        break;
    case GSM_STATE_READY:
        if (gsm_compare(gsm->lastcmd, "AT+CLIP=1")) { /* Calling Line Identification. Enable CLI presentation */
            g20_send(gsm, "AT+CLCC=1\r"); /* List Current Calls. Enable CLCC unsolicited indication */
        }
        if (gsm_compare(gsm->lastcmd, "AT+CLCC=1")) {
            g20_send(gsm, "AT+MADIGITAL=1\r"); /* Analog -> Digital Switching */
        }
        if (gsm_compare(gsm->lastcmd, "AT+MADIGITAL=1")) {
            snprintf(tmp, sizeof(tmp) - 1, "AT+MAVOL=8,1,%d\r", GSM_DIGITAL_VOLUME);
            g20_send(gsm, tmp); /* set Digital TX Voice level */
        }
        if (gsm_compare(gsm->lastcmd, "AT+MAVOL=8,1,")) {
            g20_send(gsm, "AT+MAMUT=3,1\r"); /* set Mic and Headset mic Mute*/
        }
        if (gsm_compare(gsm->lastcmd, "AT+MAMUT=3,1")) {
            g20_send(gsm, "ATS7=255\r"); /* set number of seconds to wait forf connection completion */
        }
        if (gsm_compare(gsm->lastcmd, "ATS7=255")) {
            g20_send(gsm, "AT+CLIR=0\r"); /* Calling Line Identification Restriction. Presentation indicator is used according to the subscription of the CLIR service */
        }
        if (gsm_compare(gsm->lastcmd, "AT+CLIR=0")) {
            g20_send(gsm, "AT+MAFEAT=7,0\r"); /* Disable Side tone, Echo cancel and Noise suppress */
        }
        if (gsm_compare(gsm->lastcmd, "AT+MAFEAT=7,0")) {
            g20_send(gsm, "AT+CMGF=0\r"); /* Set Message Format: PDU mode (default) */
        }
        if (gsm_compare(gsm->lastcmd, "AT+CMGF=0")) {
            g20_send(gsm, "AT+CNMI=0,2,2,0,0\r"); /* Set New Message Indications to Terminal */
        }
        
        callerid[0] = '\0';
        gsm->ev.ring.callingnum[0] = '0';
        if (gsm_compare(buf2, "+CLIP:")) { /* ring */
            gsm->state = GSM_STATE_RINGING;
            if (sscanf(buf2, "+CLIP: \"%s\",129,,128,\"\",0", callerid) == 1) {
                a = strchr(callerid, '"');
                if (a) {
                    *a = '\0';
                }
            }
            gsm->hungup = 0;
            if (!strlen(callerid)) {
                strncpy(gsm->ev.ring.callingnum, "", sizeof(gsm->ev.ring.callingnum) - 1);
            } else {
                strncpy(gsm->ev.ring.callingnum, callerid, sizeof(gsm->ev.ring.callingnum) - 1);
            }
            gsm->ev.e = GSM_EVENT_RING;
            return &gsm->ev;
        } else if (gsm_compare(buf2, "+CMT:")) { /* SMS received */
            gsm->state = GSM_STATE_SM_RECEIVING;
        } else if (gsm_compare(buf2, "NO CARRIER")) {
            if (gsm_compare(gsm->lastcmd, "ATD")) {
                gsm->state = GSM_STATE_READY;
                gsm->ev.e = GSM_EVENT_HANGUP;
                gsm->ev.hangup.cause = 16;
                return &gsm->ev;
            }
        } else if (gsm_compare(buf2, "OK")) {
            if (gsm_compare(gsm->lastcmd, "ATD")) {
                gsm->state = GSM_STATE_CALL_PRESENT;
            }
            /* sms sent */
        } else if (gsm_compare(buf2, "+CREG")) {
        } else if (gsm_compare(buf2, "AT+CLIP=1")) { 
        } else if (gsm_compare(buf2, "AT+CLCC=1")) {
        } else if (gsm_compare(buf2, "AT+MADIGITAL=1")) {
        } else if (gsm_compare(buf2, "AT+MAVOL=8,1,")) { 
        } else if (gsm_compare(buf2, "AT+MAMUT=3,1")) { 
        } else if (gsm_compare(buf2, "ATS7=255")) { 
        } else if (gsm_compare(buf2, "AT+CLIR=0")) { 
        } else if (gsm_compare(buf2, "AT+MAFEAT=7,0")) { 
        } else if (gsm_compare(buf2, "AT+CMGF=0")) { 
        } else if (gsm_compare(buf2, "AT+CNMI=0,2,2,0,0")) { 
        } else if (gsm_compare(buf2, "+CLCC:")) {
        } else if (gsm_compare(buf2, "RING")) {
        } else if (gsm_compare(buf2, "AT+CHUP")) {
            /* echo */
        } else {
            gsm_error(gsm, "!%s!\n", buf2);
        } 
        break;
    case GSM_STATE_RINGING:
        if (gsm_compare(buf2, "NO CARRIER")) { /* hang up */
            gsm->state = GSM_STATE_READY;
            gsm->ev.e = GSM_EVENT_HANGUP;
            gsm->ev.hangup.cause = 16;
            return &gsm->ev;
        } else if (gsm_compare(buf2, "OK")) {
            if (gsm_compare(gsm->lastcmd, "ATA")) { /* Answer */
                gsm->state = GSM_STATE_CALL_ACTIVE;
            }
        } else if (gsm_compare(buf2, "+CLCC: 1,0,6")) { // ?????????????????????????
            gsm->state = GSM_STATE_READY;
            if (!gsm_compare(gsm->lastcmd, "ATH")) {
                gsm->ev.e = GSM_EVENT_HANGUP;
                return &gsm->ev;
            } 
        } else if (gsm_compare(buf2, "BUSY")) { // ????????????????????????????
            gsm->state = GSM_STATE_READY;
            gsm->ev.e = GSM_EVENT_HANGUP;
            gsm->ev.hangup.cause = 17;
            return &gsm->ev;
        } else if (gsm_compare(buf2, "+CLIP:")) {
        } else if (gsm_compare(buf2, "RING")) {
        } else if (gsm_compare(buf2, "ATA")) {
        } else if (gsm_compare(buf2, "+CLCC:")) {
            /* echo */
        } else {
            gsm_error(gsm, "!%s!\n", buf2);
        } 
        break;
    case GSM_STATE_CALL_ACTIVE:
        if (gsm_compare(buf2, "NO CARRIER")) { /* hang up after answer */
            gsm->state = GSM_STATE_READY;
            gsm->ev.e = GSM_EVENT_HANGUP;
            gsm->ev.hangup.cause = 16;
            return &gsm->ev;
        } else if ((gsm_compare(buf2, "+CLCC: 1,0,6")) || (gsm_compare(buf2, "+CLCC: 1,1,6"))) {
            gsm->state = GSM_STATE_READY;
            gsm->ev.e = GSM_EVENT_HANGUP;
            gsm->ev.hangup.cause = 16;
            return &gsm->ev;
        } else if (gsm_compare(buf2, "OK")) {
            if (gsm_compare(gsm->lastcmd, "ATH")) {
                gsm->state = GSM_STATE_READY;
            }
        } else if (gsm_compare(buf2, "+CREG")) {
        } else if (gsm_compare(buf2, "+CLCC: 1,0,0")) {
        } else if (gsm_compare(buf2, "+CLCC: 1,1,0")) { /* Active */
        } else if (gsm_compare(buf2, "AT+CHUP")) { /* causes the TA to hang up the current GSM call of the g20. */
            /* echo */
            gsm_error(gsm, "!%s!\n", buf2);
        } 
        break;
    case GSM_STATE_CALL_INIT:
        if (gsm_compare(buf2, "AT+CLIR=")) {
            /* echo */
        } else if (gsm_compare(buf2, "OK")) {
            if (gsm_compare(gsm->lastcmd, "AT+CLIR=")) {
                gsm->lastcmd[0] = '\0';
                if (strlen(gsm->cpn)) {
                    memset(buf, 0x0, sizeof(buf));
                    snprintf(buf, sizeof(buf) - 1, "ATD%s;\r", gsm->cpn);
                    g20_switch_state(gsm, GSM_STATE_CALL_PRESENT, buf);
                } else {
                    g20_switch_state(gsm, GSM_STATE_READY, NULL);
                }
            }
        }
        break;
    case GSM_STATE_CALL_PRESENT:
        /* Connection Failure - NO CARRIER or BUSY or NO ANSWER */
        if (gsm_compare(buf2, "+CME ERROR")) {
            /* modem is in weird state, reboot */
            gsm_error(gsm, "error while dialing (%s)\n", buf2);
            gsm->state = GSM_STATE_ERROR;
            gsm->ev.e = GSM_EVENT_ERROR;
            gsm->ev.error.hard = 1;
            gsm->ev.error.cause = 44; /* AST_CAUSE_REQUESTED_CHAN_UNAVAIL */
            strncpy(gsm->ev.error.err, buf2, sizeof(gsm->ev.error.err) - 1);
            return &gsm->ev;
        } else if (gsm_compare(buf2, "NO CARRIER")) { /* no service */
            gsm->state = GSM_STATE_READY;
            if (strncmp(gsm->lastcmd, "ATH", 3)) {
                gsm->ev.e = GSM_EVENT_HANGUP;
                gsm->ev.hangup.cause = 16;
                return &gsm->ev;
            }
        } else if (gsm_compare(buf2, "NO ANSWER")) {
            gsm->state = GSM_STATE_READY;
            if (strncmp(gsm->lastcmd, "ATH", 3)) {
                gsm->ev.e = GSM_EVENT_HANGUP;
                gsm->ev.hangup.cause = 16;
                return &gsm->ev;
            }
        } else if (gsm_compare(buf2, "+CLCC: 1,0,6")) {
            gsm->hungup = 1;
            /* do not report the hangup event yet. wait for the cause code */
/*          gsm->state = GSM_STATE_READY;
                if (strncmp(gsm->lastcmd, "ATH", 3)) {
            gsm->ev.e = GSM_EVENT_HANGUP;
            gsm->ev.hangup.cause = 16;
            return &gsm->ev;
            } */
        } else if (gsm_compare(buf2, "+CLCC: 1,0,3")) { /* Alerting (MO call) */
            gsm->ev.e = GSM_EVENT_ALERTING;
            return &gsm->ev;
        } else if (gsm_compare(buf2, "OK")) {
            if (gsm_compare(gsm->lastcmd, "ATD")) {
                gsm->lastcmd[0] = '\0';
            } else {
                gsm->state = GSM_STATE_CALL_ACTIVE;
                gsm->ev.e = GSM_EVENT_ANSWER;
                return &gsm->ev;
            }
        } else if (gsm_compare(buf2, "BUSY")) {
            gsm->state = GSM_STATE_READY;
            gsm->ev.e = GSM_EVENT_HANGUP;
            gsm->ev.hangup.cause = 17;
            return &gsm->ev;
        } else if (gsm_compare(buf2, "ATD")) {
        } else if (gsm_compare(buf2, "+CLCC: 1,0,2")) { // Dialing (MO call)
        } else if (gsm_compare(buf2, "+CLCC: 1,0,0")) {
        } else if (gsm_compare(buf2, "+CREG")) {
        } else if (gsm_compare(buf2, "AT+CHUP")) {
        } else {
            gsm_error(gsm, "!%s!\n", buf2);
        } 
        break;
    case GSM_STATE_SM_SEND_REQ:
        if (gsm_compare(buf2, "> ")) {
            g20_transmit(gsm, gsm->msgbuf, strlen(gsm->msgbuf)); 
            memset(gsm->msgbuf, 0x0, sizeof(gsm->msgbuf));
            snprintf(gsm->msgbuf, sizeof(gsm->msgbuf) - 1, "%c", 0x1A);
            g20_transmit(gsm, gsm->msgbuf, strlen(gsm->msgbuf)); 
            gsm->state = GSM_STATE_SM_SENDING;
        } else if (gsm_compare(buf2, "+CMS ERROR")) {
            gsm_error(gsm, "error sending SMS on span %d (%s)\n", gsm->span,  buf2);
            gsm->state = GSM_STATE_READY;
        } else if (gsm_compare(buf2, "+CME ERROR")) {
            gsm_error(gsm, "error sending SMS on span %d (%s)\n", gsm->span,  buf2);
            gsm->state = GSM_STATE_READY;
        }
        break;
        case GSM_STATE_SM_SENDING:
        if (gsm_compare(buf2, "+CMGS:")) {
            g20_switch_state(gsm, GSM_STATE_READY, "AT+CMGF=0\r"); /* pdu mode (default) */
            gsm_message(gsm, "SMS sent successfully on span %d.\n", gsm->span);
        } else if (!strncmp(buf2, "+CMS ERROR", 9)) {
            gsm_error(gsm, "error sending sms (%s)\n", buf2);
            gsm->state = GSM_STATE_READY;
        }
        break;
    case GSM_STATE_SM_RECEIVING:
        if (strncmp(gsm->lastcmd, "AT+CNMA", 7)) {
            strncpy(gsm->msgbuf, buf2, sizeof(gsm->msgbuf) - 1); /* get SMS */
            g20_send(gsm, "AT+CNMA\r"); /* New Message Acknowledgment */
        }
        if (gsm_compare(buf2, "OK")) {
            gsm->lastcmd[0] = '\0';
            gsm->state = GSM_STATE_READY;
            gsm->ev.e = GSM_EVENT_SM_RECEIVED;
            if (!gsm_pdu2sm_event(gsm, gsm->msgbuf)) {
                return &gsm->ev;
            }
        }
        break;
    case GSM_STATE_HANGUP_REQ:
        if ((gsm_compare(buf2, "+CLCC: 1,0,6")) || (gsm_compare(buf2, "+CLCC: 1,1,6"))){
            /* MO released or MT released */
//      if ((gsm_compare(buf2, "NO CARRIER")) || (gsm_compare(buf2, "+CLCC: 1,0,6"))){
            gsm->state = GSM_STATE_READY;
            if (gsm->restart_timer) {
                gsm_schedule_del(gsm, gsm->restart_timer);
                gsm->restart_timer = 0;
            }
        }
        break;
        case GSM_STATE_STATUS_REQ:
        if (!strncmp(buf2, "+CSQ:", 5)){
            if (gsm->restart_timer) {
                gsm_schedule_del(gsm, gsm->restart_timer);
                gsm->restart_timer = 0;
            }
            gsm->state = GSM_STATE_READY;
            gsm_message(gsm, "signal strength on span %d (%s) [min 0, max 31]\n", gsm->span, buf2);
        }
        break;

    }
    
    return NULL;
}
