/*
 * libgsmat: An implementation of GSM 07.07
 *
 * Written by Klaus-Peter Junghanns <support@junghanns.net>
 *
 * Copyright (C) 2005, Junghanns.NET GmbH
 * All Rights Reserved.
 *
 * Parts taken from libpri-1.0.9
 * Written by Mark Spencer <markster@linux-support.net>
 *
 * Copyright (C) 2001, Linux Support Services, Inc.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */
 
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/select.h>
#include <stdarg.h>
#include <math.h>
#include "libgsmat.h"
#include "gsm_internal.h"
#include "g20.h"
#include "em200.h"
#include "sim340dz.h"
#include "m20.h"

char *gsm_modem2str(int node)
{
	switch(node) {
	case GSM_MODEM_MOTOROLA_G20:
		return "Motorola g20";
	case GSM_MODEM_SIM340DZ:
		return "SIM340DZ";
	case GSM_MODEM_EM200:
		return "EM200";
	default:
		return "Invalid value";
	}
}

char *gsm_get_span_name(struct gsm_modul *gsm)
{	
	if (!gsm)
	{
		return NULL;
	}
	
	char *idx = strchr(gsm->span_name, '/');
	if (idx !=0)
	{
		gsm->span_name[idx - gsm->span_name] = '\0';
	}
		
	return gsm->span_name;
}

int gsm_span_name_is(struct gsm_modul *gsm, char *span_name)
{	
	char *gsm_span_name = gsm_get_span_name(gsm);
	return !(strcmp(gsm_span_name, span_name));
}

void gsm_send_pin(struct gsm_modul *gsm, char *pin) {
	switch (gsm->modemtype) {
		case GSM_SWITCH_G20:
    		g20_send_pin(gsm, pin);
			break;
		case GSM_SWITCH_SIM340DZ:
    		sim340dz_send_pin(gsm, pin);
			break;
		case GSM_SWITCH_EM200:
    		em200_send_pin(gsm, pin);
			break;
		case GSM_SWITCH_M20:
    		m20_send_pin(gsm, pin);
			break;
		default:
			break;				
	}
}

struct gsm_modul *gsm_new(int fd, int modemtype, char *pin, int span, int channel)
{
	struct gsm_modul *p;
	p = malloc(sizeof(struct gsm_modul));
	if (p) {
		memset(p, 0, sizeof(struct gsm_modul));
		p->fd = fd;
		p->span = span;
		p->channel = channel;
		p->modemtype = modemtype;
		p->echo = 1;
		p->sanidx = 0;
		p->lastres = 0;
		p->dchanup = 0;
		p->restart_timer = 0;
		p->state = GSM_STATE_POWER_ON;
		p->schedev = 0;
		p->restart_timer = 0;
		
		/* Start MUX layer */
		switch (p->modemtype) {
			case GSM_SWITCH_G20:
				g20_start(p);
				break;
			case GSM_SWITCH_SIM340DZ:
				sim340dz_start(p);
				break;
			case GSM_SWITCH_EM200:
				em200_start(p);
				break;
			case GSM_SWITCH_M20:
				m20_start(p);
				break;
			default:
				return NULL;				
		}
	}
	return p;
}

int gsm_restart(struct gsm_modul *p, int when)
{
    if (p) {
	if (p->debug)
	    gsm_message(p, "scheduling restart\n");
	if (!p->restart_timer)
		switch(p->modemtype) {
			case GSM_SWITCH_G20:
	    		p->restart_timer = gsm_schedule_event(p, when, g20_restart, p);
				break;
			case GSM_SWITCH_SIM340DZ:
	    		p->restart_timer = gsm_schedule_event(p, when, sim340dz_restart, p);
				break;
			case GSM_SWITCH_EM200:
	    		p->restart_timer = gsm_schedule_event(p, when, em200_restart, p);
				break;
			case GSM_SWITCH_M20:
	    		p->restart_timer = gsm_schedule_event(p, when, m20_restart, p);
				break;
			default:
				return -1;				
			
		}
			
        return 0;
    }
	
    return -1;
}

int gsm_wait(struct gsm_modul *p)
{
    if (p) {
	while ((p->state == GSM_STATE_SM_SEND_REQ) || (p->state == GSM_STATE_SM_SENDING)) {
	    usleep(300);
	}
        return 0;
    }
    return -1;
}

int gsm_available(struct gsm_modul *p) {
    if (p->debug)
	gsm_message(p, "state = %d\n", p->state);
    if (p->state == GSM_STATE_READY) {
		return 1;
    } else {
		return 0;
    }
}

char *gsm_event2str(int id)
{
	switch(id) {
	case GSM_EVENT_DCHAN_UP:
		return "D-Channel Up";
	case GSM_EVENT_DCHAN_DOWN:
		return "D-channel Down";
	case GSM_EVENT_RING:
		return "Ring";
	default:
		return "Unknown Event";
	}
}

gsm_event *gsm_check_event(struct gsm_modul *gsm, int doread)
{
	char buf[1024];
	int res = 0;
	gsm_event *e = NULL;
	
	if (doread) {
//	    gsm_error(gsm, "check event\n");
	    res = read(gsm->fd, buf, sizeof(buf));

		if (res < 0) {
			if (errno != EAGAIN)
				gsm_error(gsm, "Read on %d failed: %s\n", gsm->fd, strerror(errno));
			return NULL;
	    }
	}
	switch(gsm->modemtype) {
		case GSM_SWITCH_G20:
			e = g20_receive(gsm, buf, res);
			break;
		case GSM_SWITCH_SIM340DZ:
			e = sim340dz_receive(gsm, buf, res);
			break;
		case GSM_SWITCH_EM200:
			e = em200_receive(gsm, buf, res);
			break;
		case GSM_SWITCH_M20:
			e = m20_receive(gsm, buf, res);
			break;
		default:
			return NULL;						
	}
	
	return e;
}

static int wait_gsm(struct gsm_modul *gsm)
{	
	struct timeval *tv, real;
	fd_set fds;
	int res;
	FD_ZERO(&fds);
	FD_SET(gsm->fd, &fds);
	tv = gsm_schedule_next(gsm);
	if (tv) {
		gettimeofday(&real, NULL);
		real.tv_sec = tv->tv_sec - real.tv_sec;
		real.tv_usec = tv->tv_usec - real.tv_usec;
		if (real.tv_usec < 0) {
			real.tv_usec += 1000000;
			real.tv_sec -= 1;
		}
		if (real.tv_sec < 0) {
			real.tv_sec = 0;
			real.tv_usec = 0;
		}
	}
	res = select(gsm->fd + 1, &fds, NULL, NULL, tv ? &real : tv);
	if (res < 0) 
		return -1;
	return res;
}

gsm_event *gsm_mkerror(struct gsm_modul *gsm, char *errstr)
{
	/* Return a configuration error */
//	gsm->ev.err.e = GSM_EVENT_CONFIG_ERR;
	strncpy(gsm->ev.error.err, errstr, sizeof(gsm->ev.error.err) - 1);
	return &gsm->ev;
}


gsm_event *gsm_dchannel_run(struct gsm_modul *gsm, int block)
{
	gsm_event *e;
	int res;
	if (!gsm)
		return NULL;
	if (block) {
		do {
			e =  NULL;
			res = wait_gsm(gsm);
			/* Check for error / interruption */
			if (res < 0) 
				return NULL;
			if (!res) {
			    e = gsm_schedule_run(gsm);
			} else {
				e = gsm_check_event(gsm,1);
			}
		} while(!e);
	} else {
		e = gsm_check_event(gsm, 1);
		return e;
	}
	return e;
}

void gsm_set_debug(struct gsm_modul *gsm, int debug)
{
	if (!gsm)
		return;
	gsm->debug = debug;
}

void gsm_set_debug_fd(struct gsm_modul *gsm, int fd)
{
	if (!gsm)
		return;
	gsm->debugfd = fd;
}

void gsm_destroycall(struct gsm_modul *gsm, gsm_call *call)
{
//	if (pri && call)
//		__q931_destroycall(pri, call);
	return;
}

int gsm_answer(struct gsm_modul *gsm)
{
	if (!gsm)
		return -1;

	switch(gsm->modemtype) {
		case GSM_SWITCH_G20:
			return g20_answer(gsm);
		case GSM_SWITCH_SIM340DZ:
			return sim340dz_answer(gsm);
		case GSM_SWITCH_EM200:
			return em200_answer(gsm);
		case GSM_SWITCH_M20:
			return m20_answer(gsm);
		default:
			break;		
	}

	return -1;
}

int gsm_hangup(struct gsm_modul *gsm)
{
	if (!gsm)
		return -1;
	switch(gsm->modemtype) {
		case GSM_SWITCH_G20:
			return g20_hangup(gsm);
		case GSM_SWITCH_SIM340DZ:
			return sim340dz_hangup(gsm);
		case GSM_SWITCH_EM200:
			return em200_hangup(gsm);
		case GSM_SWITCH_M20:
			return m20_hangup(gsm);
		default:
			break;		
	}

	return -1;
}

gsm_call *gsm_new_call(struct gsm_modul *gsm)
{
	if (!gsm)
		return NULL;
//	return q931_new_call(pri);
	return NULL;
}

void gsm_dump_event(struct gsm_modul *gsm, gsm_event *e)
{
	if (!gsm || !e)
		return;
	gsm_message(gsm, "Event type: %s (%d)\n", gsm_event2str(e->gen.e), e->gen.e);
}


int gsm_dial(struct gsm_modul *gsm, int clir, char *called)
{
	if (!gsm)
		return -1;
	
	switch(gsm->modemtype) {
		case GSM_SWITCH_G20:
			return g20_dial(gsm, clir, called);
		case GSM_SWITCH_SIM340DZ:
			return sim340dz_dial(gsm, clir, called);
		case GSM_SWITCH_EM200:
			return em200_dial(gsm, clir, called);
		case GSM_SWITCH_M20:
			return m20_dial(gsm, clir, called);
		default:
			break;		
	}

	return -1;
}	

int gsm_sms_send_text(struct gsm_modul *gsm, char *destination, char *msg) {
	if (!gsm)
		return -1;

	switch(gsm->modemtype) {
		case GSM_SWITCH_G20:
			return g20_sms_send_text(gsm, destination, msg);
		case GSM_SWITCH_SIM340DZ:
			return sim340dz_sms_send_text(gsm, destination, msg);
		case GSM_SWITCH_EM200:
			return em200_sms_send_text(gsm, destination, msg);
		case GSM_SWITCH_M20:
			return m20_sms_send_text(gsm, destination, msg);
		default:
			break;		
	}
	
	return -1;
}

int gsm_sms_send_pdu(struct gsm_modul *gsm, char *pdu) {
	if (!gsm)
		return -1;

	switch(gsm->modemtype) {
		case GSM_SWITCH_G20:
			return g20_sms_send_pdu(gsm, pdu);
		case GSM_SWITCH_SIM340DZ:
			return sim340dz_sms_send_pdu(gsm, pdu);
		case GSM_SWITCH_EM200:
//			return em200_sms_send_pdu(gsm, pdu);
			return 0;
		case GSM_SWITCH_M20:
			return m20_sms_send_pdu(gsm, pdu);
		default:
			break;		
	}

	return -1;
}

int gsm_request_status(struct gsm_modul *p) {
	if (!p)
		return -1;
	
	switch(p->modemtype) {
		case GSM_SWITCH_G20:
			if (!p->restart_timer)
			    p->restart_timer = gsm_schedule_event(p, 1000, g20_restart, p);
			return g20_request_status(p);
		case GSM_SWITCH_SIM340DZ:
			if (!p->restart_timer)
			    p->restart_timer = gsm_schedule_event(p, 1000, sim340dz_restart, p);
			return sim340dz_request_status(p);
		case GSM_SWITCH_EM200:
			if (!p->restart_timer)
			    p->restart_timer = gsm_schedule_event(p, 1000, em200_restart, p);
			return em200_request_status(p);
		case GSM_SWITCH_M20:
			if (!p->restart_timer)
			    p->restart_timer = gsm_schedule_event(p, 1000, m20_restart, p);
			return m20_request_status(p);
		default:
			break;		
	}
	
	return -1;
}

static void (*__gsm_error)(char *stuff, int span);
static void (*__gsm_message)(char *stuff, int span);

void gsm_set_message(void (*func)(char *stuff, int channel))
{
	__gsm_message = func;
}

void gsm_set_error(void (*func)(char *stuff, int channel))
{
	__gsm_error = func;
}

void gsm_message(struct gsm_modul *gsm,char *fmt, ...)
{
	char tmp[1024];
	va_list ap;
	va_start(ap, fmt);
	vsnprintf(tmp, sizeof(tmp), fmt, ap);
	va_end(ap);
	if (__gsm_message)
		__gsm_message(tmp, gsm->channel);
	else
		fputs(tmp, stdout);
}

void gsm_error(struct gsm_modul *gsm,char *fmt, ...)
{
	char tmp[1024];
	va_list ap;
	va_start(ap, fmt);
	vsnprintf(tmp, sizeof(tmp), fmt, ap);
	va_end(ap);
	if (__gsm_error)
		__gsm_error(tmp, gsm->channel);
	else
		fputs(tmp, stderr);
}

int gsm_fd(struct gsm_modul *gsm)
{
	return gsm->fd;
}

int gsm_compare( char *str_at, char *str_cmp) {
	int res;
	int i;
	int j =0;
	char buf[1024];
	int k = strlen(str_at);

	res = strncmp(str_at, str_cmp, strlen(str_cmp));
	if (!res) {
		return 1;
	}

	for (i=0; i < k ;i++) {
        if ((str_at[i] != '\r') && (str_at[i] != '\n') ) {
            buf[j++] = str_at[i];
		}
		if ((str_at[i] == '\n')) {
			buf[j++] = 0x0;
			if (j > 1) {
				res = strncmp(buf, str_cmp, strlen(str_cmp));
				if (!res) {
					return 1;
				}
			}
			j=0;
		}
    }
	return 0;
}


void gsm_set_imsi(struct gsm_modul *gsm, char *h) 
{
    char buf[64];
    int i;
    int j = 0;

    for (i=0; i < strlen(h) ;i++) {
        if ((h[i] != '\r') && (h[i] != '\n') ) {
            buf[j++] = h[i];
        }   
        if ((h[i] == '\n') && (j > 0)) {
            buf[j++] = 0x0;
            strncpy(gsm->imsi, buf, j);
            return;
        }
    }
}

void gsm_set_imei(struct gsm_modul *gsm,char *h)
{
    char buf[64];
    int i;
    int j = 0;

    for (i=0; i < strlen(h) ;i++) {
        if ((h[i] != '\r') && (h[i] != '\n') ) {
            buf[j++] = h[i];
        }   
        if ((h[i] == '\n') && (j > 0)) {
            buf[j++] = 0x0;
            strncpy (gsm->imei, buf, j);
            return;
        }
    }
}
