/*
 * libgsmat: An implementation of GSM 07.07
 *
 * Written by Klaus-Peter Junghanns <support@junghanns.net>
 *
 * Copyright (C) 2005, Junghanns.NET GmbH
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */
 
#ifndef _GSM_INTERNAL_H
#define _GSM_INTERNAL_H

#include <sys/time.h>
#include <stdlib.h>

#define DEBUGFMT  "[%s(%d)-%s]  "
#define DEBUGARGS __FILE__,__LINE__,__FUNCTION__

typedef enum gsm_state {
    POWER_OFF,
    POWER_ON,
    INIT_0,
    INIT_1,
    INIT_2,
    SIM_PIN_REQ,
    SIM_PUK_REQ,
    SIM_READY,
    READY,
    CALL_PRESENT,
    CALL_ACTIVE,
    RINGING
} gsm_state;

#define GSM_STATE_POWER_OFF              0
#define GSM_STATE_POWER_ON               (GSM_STATE_POWER_OFF + 1)
#define GSM_STATE_INIT_0                 (GSM_STATE_POWER_ON + 1)
#define GSM_STATE_INIT_1                 (GSM_STATE_INIT_0 + 1)
#define GSM_STATE_INIT_2                 (GSM_STATE_INIT_1 + 1)
#define GSM_STATE_INIT_3                 (GSM_STATE_INIT_2 + 1)
#define GSM_STATE_DISABLE_ECHO           (GSM_STATE_INIT_3 + 1)
#define GSM_STATE_VERSION_REQ            (GSM_STATE_DISABLE_ECHO + 1)
#define GSM_STATE_IMSI_REQ               (GSM_STATE_VERSION_REQ + 1)
#define GSM_STATE_IMEI_REQ               (GSM_STATE_IMSI_REQ + 1)
#define GSM_STATE_REPORT_ERROR           (GSM_STATE_IMEI_REQ + 1)
#define GSM_STATE_CHECK_SIM_SECURITY     (GSM_STATE_REPORT_ERROR + 1)
#define GSM_STATE_NET_REQ                (GSM_STATE_CHECK_SIM_SECURITY + 1)
#define GSM_STATE_SIM_PIN_REQ            (GSM_STATE_NET_REQ + 1)
#define GSM_STATE_SIM_PUK_REQ            (GSM_STATE_SIM_PIN_REQ + 1)
#define GSM_STATE_SIM_READY              (GSM_STATE_SIM_PUK_REQ + 1)
#define GSM_STATE_READY                  (GSM_STATE_SIM_READY + 1)
#define GSM_STATE_CALL_INIT              (GSM_STATE_READY + 1)
#define GSM_STATE_CALL_PRESENT           (GSM_STATE_CALL_INIT + 1)
#define GSM_STATE_CALL_ACTIVE            (GSM_STATE_CALL_PRESENT + 1)
#define GSM_STATE_RINGING                (GSM_STATE_CALL_ACTIVE + 1)

#define GSM_STATE_SM_SEND_REQ            (GSM_STATE_RINGING + 1)
#define GSM_STATE_SM_SENDING             (GSM_STATE_SM_SEND_REQ + 1)
#define GSM_STATE_SM_RECEIVING           (GSM_STATE_SM_SENDING + 1)

#define GSM_STATE_HANGUP_REQ             (GSM_STATE_SM_RECEIVING + 1)
#define GSM_STATE_STATUS_REQ             (GSM_STATE_HANGUP_REQ + 1)
#define GSM_STATE_ERROR                  (GSM_STATE_STATUS_REQ + 1)
#define GSM_STATE_ERROR_HARD             (GSM_STATE_ERROR + 1)

#define GSM_DIGITAL_VOLUME  5

#define GSM_SMS_TP_RP		128	// (1000 0000) Reply path. Parameter indicating that reply path exists.
#define GSM_SMS_TP_UDHI		64	// (0100 0000) User data header indicator. This bit is set to 1 if the User Data field starts with a header.
#define GSM_SMS_TP_SRI		32	// (0010 0000) Status report indication. This bit is set to 1 if a status report is going to be returned to the SME
#define GSM_SMS_TP_MMS		4	// (0000 0100) More messages to send. This bit is set to 0 if there are more messages to send
#define GSM_SMS_TP_MTI		3	// (0000 0011) Message type indicator. See GSM 03.40 subsection 9.2.3.1

struct gsm_sched {
    struct timeval when;
    void (*callback)(void *data);
    void (*callback2)(void *data, int);
    void *data;
    char hasdata2;
    int data2;
};

/* No more than 128 scheduled events */
/* [1, 128) */
#define MAX_SCHED 128

struct gsm_modul {
    int fd;             /* File descriptor for D-Channel */
    int span;           /* zaptel span */
    int channel;            /* "d" channel */
    struct gsm_sched gsm_sched[MAX_SCHED];  /* Scheduled events */
    int debug;          /* Debug stuff */
    int debugfd;
    int modemtype;      /* Switch type */
    
    char dchanup;

    char clip[256];
    char cpn[256];
    int clir;
    int hungup;
    char lastcmd[1024];
    char lasterr[1024];
    char sanbuf[4096];
    char msgbuf[4096];
    int smslen;
    int sanidx;
    int echo;
    int lastres;
    /* Used by scheduler */
    struct timeval tv;
    int schedev;
    int state;
    int sim_state;
    int restart_timer;
    gsm_event ev;       /* Static event thingy */

    char span_name[20];

    char imei[64];              /* span imei */
    char imsi[64];              /* sim imsi */
    char net_name[64];          /* Network friendly name */
    int network;            /* 0 unregistered - 1 home - 2 roaming */
    int coverage;           /* net coverage -1 not signal */
    int status;
};

/*
 * from copy_string.c
 */
 
void gsm_copy_string(char *dst, const char *src, size_t size);

extern int gsm_schedule_event(struct gsm_modul *gsm, int ms, void (*function)(void *data), void *data);
extern int gsm_schedule_event2(struct gsm_modul *gsm, int ms, void (*function)(void *data, int data2), void *data, int data2);

extern gsm_event *gsm_schedule_run(struct gsm_modul *gsm);

extern void gsm_schedule_del(struct gsm_modul *gsm, int ev);

extern gsm_event *gsm_mkerror(struct gsm_modul *gsm, char *errstr);

extern void gsm_message(struct gsm_modul *gsm, char *fmt, ...);

extern void gsm_error(struct gsm_modul *gsm, char *fmt, ...);

extern unsigned long gsm_hex2int(char *hex, int len);

extern int gsm_pdu2sm_event(struct gsm_modul *gsm, char *pdu);

extern int sms_get_str(char *in, size_t inlen, char *out, size_t outlen);

extern int sms_set_str(char *in, char *out);

extern int gsm_text2sm_event(struct gsm_modul *gsm, char *sms_info);

#endif

